#include <stdio.h>
#include <errno.h>
#include <math.h>
#ifdef MPI
#include <mpi.h>
#endif
#include "DEFS.h"
#include "PATH.h"
#include "MC.h"
#include "PROCESS.h"
#include "CALND.h"
#include "MEM.h"
#include "EU_OPT.h"
#include "_eu_opt.h"

static REAL *CurrentPath = NULL;

#ifdef MPI
  extern int NumberofProcessors; 
  extern int MyProcessorNumber;
#endif

int   EuropeanOptionServer
                         (
                         char *UNDERLYING_descriptor,
                         char *MC_descriptor,
                         char *PROCESS_descriptor,
                         char *EU_OPTION_descriptor,
                         char *OptionContainer
                         )
{
  int err = 0;
  int status = 0;
  PATH *Underlying = NULL;
  FILE *pfile = NULL;
  int NumRates; 
  int NumTimeSteps; 
  int PeriodLength;
  int StartDate; 
  int EndDate;
  char UnderlyingContainer[MAX_FILE_NAME_LENGTH+1];

  int Lag = 0;
  PROCESS *ProcessParams = NULL;
  int Naccepted = -100;
  int NumInitPrices;
  int iopt;

  EU_OPTION **Option = NULL;


#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif
  if((pfile = fopen(UNDERLYING_descriptor,"r")) == NULL) {
        fprintf(stderr, "EuropeanOptionServer(): can't open underlying descriptor\n");
        exit(1);
  }
#ifdef MPI
  }
#endif

#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif

    getline(pfile);
    fscanf(pfile,"%d\n",&NumRates);

    getline(pfile);
    fscanf(pfile,"%d\n",&StartDate);

    getline(pfile);
    fscanf(pfile,"%d\n",&EndDate);

    getline(pfile);
    fscanf(pfile,"%d\n",&PeriodLength);
    NumTimeSteps = NumberOfPeriods(StartDate,EndDate,PeriodLength);

    getline(pfile);
    fscanf(pfile,"%s\n",UnderlyingContainer);

    fclose(pfile);

#ifdef MPI
  }
#endif

#ifdef MPI
  MPI_Bcast(&NumRates, 1, MPI_INT, MASTER, MPI_COMM_WORLD);
  MPI_Bcast(&NumTimeSteps, 1, MPI_INT, MASTER, MPI_COMM_WORLD);
#endif


  if( (Underlying = CreatePath
                             (
                             NumTimeSteps,
                             NumRates,
                             Lag=0
                             ) 
      ) == NULL)  {
      fprintf(stderr,"EuropeanOptionServer(): Could not create path for underlying\n");
      exit(1);
  }
  NumInitPrices = NumRates;

  if( (ProcessParams = SetProcessParams
                                      (
                                      NumTimeSteps,
                                      NumRates,
                                      Underlying,
                                      PROCESS_descriptor)
      )== NULL ) {
      fprintf(stderr,"EuropeanOptionServer(): Could not set underlying process parameters\n");
      exit(1);
  }

  if(err = SetupMetropolisMCSimulation(MC_descriptor)) {
     fprintf(stderr,"EuropeanOptionServer(): Metropolis set up failed\n");
     exit(1);
  }

  Option = (EU_OPTION **) mymalloc((NumInitPrices+1)*sizeof(EU_OPTION *), "EuropeanOptionServer(): Option\n"); 
  for(iopt=1; iopt <= NumInitPrices; iopt++) {
    if((Option[iopt] = CreateEuropeanOption
                                          (
                                          Underlying,
                                          EU_OPTION_descriptor
                                          )) == NULL ) {
        fprintf(stderr,"EuropeanOptionServerDemo(): Could not create option # %d\n", iopt);
        exit(1);
    }
  }

  printf("created options\n");

  Naccepted = EuropeanValuation
                              (
                              Underlying,
                              ProcessParams, 
                              Option
                              );
  status = Externalize
                     (
                     NumInitPrices,
                     Option,
                     OptionContainer
                     );

/*
  PROCESS_Free(ProcessParams,NumRates);
  for(iopt=1;iopt<=NumInitPrices;iopt++) EU_OPTION_Free(Option[iopt]);

  free((void **) Option);
  PATH_Free(Underlying);
*/
  return status;

}


int EuropeanValuation
                    (
                    PATH *Underlying, 
                    PROCESS *ProcessParams,
                    EU_OPTION **Option
	            )
    
{
  int mb; 
  int mcs;
  int  Naccepted=0;
  int size = 0;
  int NumRates = HowManyRates(Underlying);
  int NumMeasBlocks   = HowManyMeasBlocks();
  int NumRelaxBlocks  = HowManyRelaxBlocks();
  int MeasBlockSteps  = HowBigMeasBlock();
  int iopt;
  int NumInitPrices = HowManyRates(Underlying);
  REAL *Variance = GetVariances(ProcessParams,NumInitPrices);

  SetAllPathsToZero(Underlying);  
  size = (HowManyTimeSteps(Underlying) + 1)*sizeof(REAL); 
  CurrentPath = (REAL *) mymalloc(size,"EuropeanValuation(): CurrentPath\n");
  for(iopt=1;iopt<=NumInitPrices;iopt++) InitOption(Option[iopt],Variance);
  InitMom(Underlying); 

  SetRelaxFlag(NOT_RELAXED);

  if(!IsRelaxed()) for(mb=1;mb<=NumRelaxBlocks; mb++)
            Naccepted += MetropolisUpdate
                                        (
                                        MeasBlockSteps, 
                                        Underlying, 
                                        ProcessParams
                                        );
  SetRelaxFlag(RELAXED);
  while(!ImDone()) {
    for(mb=1;mb<=NumMeasBlocks; mb++) {
      Naccepted += MetropolisUpdate
                                  (
                                  MeasBlockSteps,  
                                  Underlying, 
                                  ProcessParams
                                  );
      MeasMom6(Underlying); 
      for(iopt=1;iopt<=NumInitPrices;iopt++)  MeasOption
                                                       (
                                                       iopt,
                                                       Option[iopt],
                                                       Underlying,
                                                       ProcessParams
                                                       ); 
    }
    CycleDone();
  }

  PrintMoments(); 
  for(iopt=1;iopt<=NumInitPrices;iopt++) Discount(Option[iopt]);   
  for(iopt=1;iopt<=NumInitPrices;iopt++) Statistics(Option[iopt]);
  for(iopt=1;iopt<=NumInitPrices;iopt++) PrintOptionValues(Option[iopt]);


  V_Free((void *) CurrentPath);
  V_Free((void *) Variance);
  return Naccepted;

} 



EU_OPTION *CreateEuropeanOption
                              (
                              PATH *Security,
                              char *European_option_descriptor
                              )         
{
  int err = 0;
  int size = 0;
  int NumTimeSteps = HowManyTimeSteps(Security);
  int numDoF = HowManyRates(Security);
  int dof=0;
  int nt = 0;
  int nov = 0;
  int i;
  int it;
  int iit;
  int ir;
  int iir;
  REAL rate = 0.;
  REAL price = 0.;
  int PayoffType = 0;
  int nipp = 0;
  int nfpp = 0;
  int ip;
  int iip;
  int ipar = 0;
  REAL fpar = 0.;

  FILE *dfile = NULL;
  EU_OPTION *Option = NULL;
  

  Option= (EU_OPTION *) mymalloc(sizeof(EU_OPTION), "CreateEuropeanOption(): Option\n");
  Option->Underlying = (void *) Security;
  Option->TimeGridSize = NumTimeSteps;

#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif

  if((dfile = fopen(European_option_descriptor,"r")) == NULL) {
        fprintf(stderr, "CreateEuropeanOption(): can't open description file\n");
        exit(1);
  }

  Option->ValuationTimeIndex = NULL;
  getline(dfile);
  fscanf(dfile, "%d\n",&nt);
  if(nt > NumTimeSteps || nt < 1) nt = NumTimeSteps;

#ifdef MPI
  MPI_Bcast(&nt, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);
#endif

  Option->NumOptionTimeValues = nt;
  if(nt != NumTimeSteps) {
    getline(dfile);
    size = (nt+1)*sizeof(int);
    Option->ValuationTimeIndex = (int *) 
         mymalloc(size,"CreateEuropeanOption(): ValuationTimeIndex\n");
    for(i=0;i<=nt;i++) fscanf(dfile,"%d",Option->ValuationTimeIndex+i);
#ifdef MPI
    MPI_Bcast(Option->ValuationTimeIndex, (nt+1) , MPI_INT, MASTER, MPI_COMM_WORLD);
#endif
  }

  getline(dfile);
  fscanf(dfile, "%d\n",&dof);
  if(dof > numDoF || dof < 1) {
        dof = numDoF;
  }
#ifdef MPI
  MPI_Bcast(&dof, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);

#endif

  Option->DimOfUnderlying = dof;
  size = (dof + 1)*sizeof(REAL);
  Option->InitPrice = (REAL *) mymalloc(size,"CreateEuropeanOption():InitPrice\n");

  size = (dof + 1)*sizeof(REAL *);
  Option->LogPrice = (REAL **) mymalloc(size,"CreateEuropeanOption():LogPrice\n");
  size = (NumTimeSteps + 1)*sizeof(REAL);
  for(ir=0;ir<=dof;ir++) 
    Option->LogPrice[ir] = (REAL *) mymalloc(size,"CreateEuropeanOption():LogPrice\n");

  getline(dfile);
  fscanf(dfile,"%d\n", &nov);
  Option->DimOfOption = nov;
#ifdef MPI
  MPI_Bcast(&nov, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);

#endif



  getline(dfile);
  for(ir=1;ir<=dof;ir++) {
      fscanf(dfile,"%d %f\n", &iir, &price);
      *(Option->InitPrice + iir) = price;
  }
#ifdef MPI
  MPI_Bcast(Option->InitPrice, (dof+1) , MPI_REAL, MASTER, MPI_COMM_WORLD);
#endif


  size = (NumTimeSteps+1)*sizeof(REAL);
  Option->InterestRate = (REAL *) mymalloc(size,"CreateEuropeanOption():InterestRate\n");
  getline(dfile);
  fscanf(dfile,"%d %f\n", &iit, &rate);
  for(it=0;it<=NumTimeSteps;it++) {
      *(Option->InterestRate + it) = rate;
  }
#ifdef MPI
  MPI_Bcast(Option->InterestRate, (NumTimeSteps+1) , MPI_REAL, MASTER, MPI_COMM_WORLD);
#endif
  size = (NumTimeSteps + 1)*sizeof(REAL);
  Option->Drift = (REAL *) mymalloc(size,"CreateEuropeanOption():Drift\n");


  getline(dfile);
  fscanf(dfile,"%d\n", &PayoffType);
  Option->PayoffFunction = NULL;
#ifdef MPI
  MPI_Bcast(&PayoffType, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);
#endif


  getline(dfile);
  fscanf(dfile,"%d\n",&nipp);
#ifdef MPI
  MPI_Bcast(&nipp, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);
#endif
  Option->NumIntPayoffParams = nipp;

  Option->IntPayoffParams = NULL;
  if(nipp > 0) {
    size = (nipp + 1)*sizeof(int);
    Option->IntPayoffParams = (int *) mymalloc(size,"CreateEuropeanOption():IntPayoffParams\n");

    getline(dfile);
    for(ip=1;ip<=nipp;ip++) {
        fscanf(dfile,"%d %d\n", &iip, &ipar);
        *(Option->IntPayoffParams + iip) = ipar;
    }
#ifdef MPI
    MPI_Bcast(Option->IntPayoffParams, (nipp+1) , MPI_INT, MASTER, MPI_COMM_WORLD);
#endif
  }


  getline(dfile);
  fscanf(dfile,"%d\n",&nfpp);
#ifdef MPI
  MPI_Bcast(&nfpp, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);
#endif
  Option->NumFloatPayoffParams = nfpp;

  size = (nfpp + 1)*sizeof(REAL);
  Option->FloatPayoffParams = (REAL *) mymalloc(size,"CreateEuropeanOption():FloatPayoffParams\n");

  getline(dfile);
  for(ip=1;ip<=nfpp;ip++) {
      fscanf(dfile,"%d %f\n", &iip, &fpar);
      *(Option->FloatPayoffParams + iip) = fpar;
  }
#ifdef MPI
  MPI_Bcast(Option->FloatPayoffParams, (nfpp+1) , MPI_REAL, MASTER, MPI_COMM_WORLD);
#endif
  Option->Strike = Option->FloatPayoffParams;


#ifdef MPI
  } else {

  Option->ValuationTimeIndex = NULL;
  MPI_Bcast(&nt, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);

  Option->NumOptionTimeValues = nt;
  if(nt != NumTimeSteps) {
    size = (nt+1)*sizeof(int);
    Option->ValuationTimeIndex = (int *)
         mymalloc(size,"CreateEuropeanOption(): ValuationTimeIndex\n");
    MPI_Bcast(Option->ValuationTimeIndex, (nt+1) , MPI_INT, MASTER, MPI_COMM_WORLD);
  }

  MPI_Bcast(&dof, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);
  Option->DimOfUnderlying = dof;

  size = (dof + 1)*sizeof(REAL);
  Option->InitPrice = (REAL *) mymalloc(size,"CreateEuropeanOption():InitPrice\n");

  size = (dof + 1)*sizeof(REAL *);
  Option->LogPrice = (REAL **) mymalloc(size,"CreateEuropeanOption():LogPrice\n");
  size = (NumTimeSteps + 1)*sizeof(REAL);
  for(ir=0;ir<=dof;ir++) 
    Option->LogPrice[ir] = (REAL *) mymalloc(size,"CreateEuropeanOption():LogPrice\n");

  MPI_Bcast(&nov, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);
  Option->DimOfOption = nov;

  MPI_Bcast(Option->InitPrice, (dof+1) , MPI_REAL, MASTER, MPI_COMM_WORLD);

  size = (NumTimeSteps+1)*sizeof(REAL);
  Option->InterestRate = (REAL *) mymalloc(size,"CreateEuropeanOption():InterestRate\n");
  MPI_Bcast(Option->InterestRate, (NumTimeSteps+1) , MPI_REAL, MASTER, MPI_COMM_WORLD);
  size = (NumTimeSteps + 1)*sizeof(REAL);
  Option->Drift = (REAL *) mymalloc(size,"CreateEuropeanOption():Drift\n");

  MPI_Bcast(&PayoffType, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);
/*
  Option->PayoffFunction = SetPayoffFunction(PayoffType);
*/
  Option->PayoffFunction = NULL;

  MPI_Bcast(&nipp, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);
  Option->NumIntPayoffParams = nipp;

  Option->IntPayoffParams = NULL;
  if(nipp > 0) {
    size = (nipp + 1)*sizeof(int);
    Option->IntPayoffParams = (int *) mymalloc(size,"CreateEuropeanOption():IntPayoffParams\n");
    MPI_Bcast(Option->IntPayoffParams, (nipp+1) , MPI_INT, MASTER, MPI_COMM_WORLD);
  }

  MPI_Bcast(&nfpp, 1 , MPI_INT, MASTER, MPI_COMM_WORLD);
  Option->NumFloatPayoffParams = nfpp;


  size = (nfpp + 1)*sizeof(REAL);
  Option->FloatPayoffParams = (REAL *) mymalloc(size,"CreateEuropeanOption():FloatPayoffParams\n");
  MPI_Bcast(Option->FloatPayoffParams, (nfpp+1) , MPI_REAL, MASTER, MPI_COMM_WORLD);

  Option->Strike = Option->FloatPayoffParams;
  }
#endif

  size = (NumTimeSteps + 1)*sizeof(REAL);
  Option->DiscountBeg = (REAL *) mymalloc(size, "CreateEuropeanOption():DiscountBeg\n");
  Option->DiscountEnd = (REAL *) mymalloc(size, "CreateEuropeanOption():DiscountEnd\n");

  size = (Option->DimOfOption+1) * sizeof(REAL *);

  Option->CallValue = (REAL **) mymalloc(size, "CreateEuropeanOption():Call Value\n");
  Option->PutValue = (REAL **) mymalloc(size, "CreateEuropeanOption():Put Value\n");
  Option->CallError = (REAL **) mymalloc(size, "CreateEuropeanOption():Call Error\n");
  Option->PutError = (REAL **) mymalloc(size, "CreateEuropeanOption():Put Error\n");

  Option->CallDelta = (REAL **) mymalloc(size, "CreateEuropeanOption():Call Delta\n");
  Option->PutDelta = (REAL **) mymalloc(size, "CreateEuropeanOption():Put Delta\n");
  Option->CallDeltaError = (REAL **) mymalloc(size, "CreateEuropeanOption():Call Delta Error\n");
  Option->PutDeltaError = (REAL **) mymalloc(size, "CreateEuropeanOption():Put Delta Error\n");

  Option->CallVega = (REAL **) mymalloc(size, "CreateEuropeanOption():Call Vega\n");
  Option->PutVega = (REAL **) mymalloc(size, "CreateEuropeanOption():Put Vega\n");
  Option->CallVegaError = (REAL **) mymalloc(size, "CreateEuropeanOption():Call Vega Error\n");
  Option->PutVegaError = (REAL **) mymalloc(size, "CreateEuropeanOption():Put Vega Error\n");

  Option->CallRho = (REAL **) mymalloc(size, "CreateEuropeanOption():Call Rho\n");
  Option->PutRho = (REAL **) mymalloc(size, "CreateEuropeanOption():Put Rho\n");
  Option->CallRhoError = (REAL **) mymalloc(size, "CreateEuropeanOption():Call Rho Error\n");
  Option->PutRhoError = (REAL **) mymalloc(size, "CreateEuropeanOption():Put Rho Error\n");


  size = (Option->NumOptionTimeValues + 1)*sizeof(REAL);
  for(ip=0;ip <= Option->DimOfOption; ip++) {

     Option->CallValue[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Call\n");
     Option->PutValue[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Put\n");
     Option->CallError[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Call errors\n");
     Option->PutError[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Put errors\n");

     Option->CallDelta[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Call Deltas\n");
     Option->PutDelta[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Put Deltas\n");
     Option->CallDeltaError[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Call Delta Errors\n");
     Option->PutDeltaError[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Put Delta Errors\n");

     Option->CallVega[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Call Vegas\n");
     Option->PutVega[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Put Vegas\n");
     Option->CallVegaError[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Call Vega Errors\n");
     Option->PutVegaError[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Put Vega Errors\n");

     Option->CallRho[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Call Rhos\n");
     Option->PutRho[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Put Rhos\n");
     Option->CallRhoError[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Call Rho Errors\n");
     Option->PutRhoError[ip] = (REAL *) mymalloc(size, "CreateEuropeanOption():Put Rho Errors\n");
  }

         
  return Option;

}


void InitOption
              (
              EU_OPTION *Option,
              REAL *Var
              )

{
  int ip;
  int it;
  int ir;
  REAL fac = 0.;
  REAL drift = 0.;
  int nov = Option->DimOfOption;
  int NumTimeSteps = Option->TimeGridSize;

  for(ip=0;ip <= nov; ip++) 
    for(it=0; it<= Option->NumOptionTimeValues; it++) {

      *(Option->CallValue[ip] + it) = 0.;
      *(Option->PutValue[ip] + it) = 0.;
      *(Option->CallError[ip] + it) = 0.;
      *(Option->PutError[ip] + it) = 0.;

      *(Option->CallDelta[ip] + it) = 0.;
      *(Option->PutDelta[ip] + it) = 0.;
      *(Option->CallDeltaError[ip] + it) = 0.;
      *(Option->PutDeltaError[ip] + it) = 0.;

      *(Option->CallVega[ip] + it) = 0.;
      *(Option->PutVega[ip] + it) = 0.;
      *(Option->CallVegaError[ip] + it) = 0.;
      *(Option->PutVegaError[ip] + it) = 0.;

      *(Option->CallRho[ip] + it) = 0.;
      *(Option->PutRho[ip] + it) = 0.;
      *(Option->CallRhoError[ip] + it) = 0.;
      *(Option->PutRhoError[ip] + it) = 0.;


    }

  drift = 0.;
  *(Option->Drift+0) = drift;
/*
  if(Var[1] != Var[2]) {
    fprintf(stderr,"This version does not support multiple underlyings, just multiple initial prices!\n");
    exit(1);
  }
*/
  for(it=1; it <= NumTimeSteps; it++) {
      drift += *(Option->InterestRate+it) - 0.5*Var[1]*Var[1];
      *(Option->Drift + it) = drift;
  }

  for(ir=1; ir <= Option->DimOfUnderlying; ir++) 
    for(it=0; it<= NumTimeSteps; it++) {
        *(Option->LogPrice[ir] + it) = 
        log(*(Option->InitPrice + ir)) + (*(Option->Drift + it));
    }

  fac = 1.;
  *(Option->DiscountBeg+0) = 1.;
  for(it=1; it<= NumTimeSteps; it++) {
    *(Option->DiscountBeg+it) = exp(- (*(Option->InterestRate+it))*it);
  }

  fac = 1.;
  *(Option->DiscountEnd+NumTimeSteps) = 1.;
  for(it=NumTimeSteps-1; it > 0; it--) {
  *(Option->DiscountEnd+it) = exp(- (*(Option->InterestRate+it))*(NumTimeSteps-it));
  }

}

void Discount
            (
            EU_OPTION *Option
            )
{
  int it;
  int ir;
  int NumTimeSteps = Option->TimeGridSize;
  int nov = Option->DimOfOption;
  REAL fac;

  for(ir=1; ir<=nov; ir++) {
    for(it=1; it<= NumTimeSteps; it++) {

      fac = *(Option->DiscountBeg+it);

      (*(Option->CallValue[ir]+it)) *= fac;
      (*(Option->PutValue[ir]+it)) *= fac;
      (*(Option->CallError[ir]+it)) *= fac*fac;
      (*(Option->PutError[ir]+it)) *= fac*fac;

      (*(Option->CallDelta[ir]+it)) *= fac;
      (*(Option->PutDelta[ir]+it)) *= fac;
      (*(Option->CallDeltaError[ir]+it)) *= fac*fac;
      (*(Option->PutDeltaError[ir]+it)) *= fac*fac;

      (*(Option->CallVega[ir]+it)) *= fac;
      (*(Option->PutVega[ir]+it)) *= fac;
      (*(Option->CallVegaError[ir]+it)) *= fac*fac;
      (*(Option->PutVegaError[ir]+it)) *= fac*fac;

      (*(Option->CallRho[ir]+it)) *= fac;
      (*(Option->PutRho[ir]+it)) *= fac;
      (*(Option->CallRhoError[ir]+it)) *= fac*fac;
      (*(Option->PutRhoError[ir]+it)) *= fac*fac;
    }
  }
}


void Statistics
                 (
                 EU_OPTION *Option
                 )
{
  int it;
  int ir;
  int NumTimeSteps = Option->TimeGridSize;
  int nov = Option->DimOfOption;
  int mb = HowManyMeasBlocks();
  REAL call; 
  REAL put;
  REAL call2;
  REAL put2;
  REAL call_delta; 
  REAL put_delta;
  REAL call_delta2;
  REAL put_delta2;
  REAL call_vega; 
  REAL put_vega;
  REAL call_vega2;
  REAL put_vega2;
  REAL call_rho; 
  REAL put_rho;
  REAL call_rho2;
  REAL put_rho2;
  int sfac = 2;  /* symmetry factor to count + and - paths */
  int nmeas;


  for(ir=1; ir<=nov; ir++) {

    it=0;
    nmeas = mb;
    call = (*(Option->CallValue[ir]+it))/nmeas; 
    (*(Option->CallValue[ir]+it)) = call;
    put  = (*(Option->PutValue[ir]+it))/nmeas;
    (*(Option->PutValue[ir]+it)) = put;
    (*(Option->CallError[ir]+it)) = 0.0;
    (*(Option->PutError[ir]+it)) = 0.0;

    (*(Option->CallDelta[ir]+it)) = 1.0;
    (*(Option->PutDelta[ir]+it)) = -1.0;
    (*(Option->CallDeltaError[ir]+it)) = 0.0;
    (*(Option->PutDeltaError[ir]+it)) = 0.0;

    call_vega = (*(Option->CallVega[ir]+it))/nmeas;
    (*(Option->CallVega[ir]+it)) = 0.0;
    put_vega  = (*(Option->PutVega[ir]+it))/nmeas;
    (*(Option->PutVega[ir]+it)) = 0.0;
    (*(Option->CallVegaError[ir]+it)) = 0.0;
    (*(Option->PutVegaError[ir]+it)) = 0.0;

    (*(Option->CallRho[ir]+it)) = 0.0;
    (*(Option->PutRho[ir]+it)) = 0.0;
    (*(Option->CallRhoError[ir]+it)) = 0.0;
    (*(Option->PutRhoError[ir]+it)) = 0.0;


    for(it=1; it<= NumTimeSteps; it++) {
      nmeas = sfac*mb;
      call = (*(Option->CallValue[ir]+it))/nmeas;
      (*(Option->CallValue[ir]+it)) = call;
      put  = (*(Option->PutValue[ir]+it))/nmeas;
      (*(Option->PutValue[ir]+it)) = put;
      call2 = (*(Option->CallError[ir]+it))/nmeas;
      put2  = (*(Option->PutError[ir]+it))/nmeas;
      (*(Option->CallError[ir]+it)) = sqrt((call2-call*call)/nmeas);
      (*(Option->PutError[ir]+it)) = sqrt((put2-put*put)/nmeas);

      call_delta = (*(Option->CallDelta[ir]+it))/nmeas;
      (*(Option->CallDelta[ir]+it)) = call_delta;
      put_delta  = (*(Option->PutDelta[ir]+it))/nmeas;
      (*(Option->PutDelta[ir]+it)) = put_delta;
      call_delta2 = (*(Option->CallDeltaError[ir]+it))/nmeas;
      put_delta2  = (*(Option->PutDeltaError[ir]+it))/nmeas;
      (*(Option->CallDeltaError[ir]+it)) = sqrt((call_delta2-call_delta*call_delta)/nmeas);
      (*(Option->PutDeltaError[ir]+it)) = sqrt((put_delta2-put_delta*put_delta)/nmeas);

      call_vega = 0.01*(*(Option->CallVega[ir]+it))/nmeas;
      (*(Option->CallVega[ir]+it)) = call_vega;
      put_vega  = 0.01*(*(Option->PutVega[ir]+it))/nmeas;
      (*(Option->PutVega[ir]+it)) = put_vega;
      call_vega2 = 0.0001*(*(Option->CallVegaError[ir]+it))/nmeas;
      put_vega2  = 0.0001*(*(Option->PutVegaError[ir]+it))/nmeas;
      (*(Option->CallVegaError[ir]+it)) = sqrt((call_vega2-call_vega*call_vega)/nmeas);
      (*(Option->PutVegaError[ir]+it)) = sqrt((put_vega2-put_vega*put_vega)/nmeas);

      call_rho = (*(Option->CallRho[ir]+it))/nmeas;
      (*(Option->CallRho[ir]+it)) = call_rho;
      put_rho  = (*(Option->PutRho[ir]+it))/nmeas;
      (*(Option->PutRho[ir]+it)) = put_rho;
      call_rho2 = (*(Option->CallRhoError[ir]+it))/nmeas;
      put_rho2  = (*(Option->PutRhoError[ir]+it))/nmeas;
      (*(Option->CallRhoError[ir]+it)) = sqrt((call_rho2-call_rho*call_rho)/nmeas);
      (*(Option->PutRhoError[ir]+it)) = sqrt((put_rho2-put_rho*put_rho)/nmeas);

    }

  }
/*
#ifdef MPI
  GatherResults(Option); 
#endif
*/

}


void MeasOption
              (
              int which_one,
              EU_OPTION *Option,
              PATH *Underlying,
              PROCESS *Process
              )
{
  
  int NumRates=HowManyRates(Underlying);
  int NumTimeSteps=HowManyTimeSteps(Underlying);
  int nt = Option->NumOptionTimeValues;
  REAL *StrikePrice = Option->Strike;
  int ir; 
  int it;
  int itv;
  REAL logarithmPrice;
  REAL PriceWithDrift;
  REAL del;
  int ix;
  int NumStrikes = Option->DimOfOption;
  REAL *Var = GetVariances(Process,NumRates);
  ProbFunc *pf = GetProbabilityFunction(Process);
  REAL fdelta1;
  REAL fdelta2;
  REAL fac;
  REAL f1;
  REAL f2;
  REAL s0;
  REAL delf;
  REAL final=0.0;
  REAL rho = 0.0;
 
  ir = which_one;
  SetCurrentRateIndex(Underlying, ir);
  GetCurrentPath(Underlying, CurrentPath);  
  s0 = *(Option->InitPrice+ir);
  f1 = CurrentPath[1] - CurrentPath[0];
  f2 = -CurrentPath[1] - CurrentPath[0];


  if(pf == GeneralizedCauchy) {
          fac = -8./(5.0*s0*Var[ir]*Var[ir]);
          fdelta1 = fac * f1 /(1. +f1*f1/(5.0*Var[ir]*Var[ir]));
          fdelta2 = fac * f2 /(1. +f2*f2/(5.0*Var[ir]*Var[ir]));
  } else {
      /* assuming Gaussian */
          fac = 1./(s0*Var[ir]*Var[ir]);
          fdelta1 = fac * f1;
          fdelta2 = fac * f2;
  }

  for(ix=1;ix<=NumStrikes;ix++) {

/* treat t=0 case separately */

      del= *(Option->InitPrice+ir) - StrikePrice[ix];
      if( del >= 0.0){
            *(Option->CallValue[ix]+0) +=  del;
            *(Option->CallError[ix]+0) +=  del*del;
      } else {
           *(Option->PutValue[ix]+0) += -del ;
           *(Option->PutError[ix]+0) += del*del ;
      }
  

      for(it=1; it<=nt; it++){
  
        logarithmPrice = CurrentPath[it] + *(Option->LogPrice[ir] + it);
       
        PriceWithDrift = exp( logarithmPrice );
  
        del = (PriceWithDrift- StrikePrice[ix]);
        final = CurrentPath[it] - CurrentPath[0] - Var[ir]*Var[ir]*it;
        final *= (PriceWithDrift/Var[ir]);
        rho = 0.01*StrikePrice[ix]*(REAL)it;
  
         if( del >= 0.0){
	    *(Option->CallValue[ix]+it) +=  del;
            *(Option->CallError[ix]+it) +=  del*del;
            delf = del*fdelta1;
            *(Option->CallDelta[ix]+it) +=  delf;
            *(Option->CallDeltaError[ix]+it) +=  delf*delf;
            *(Option->CallVega[ix]+it) +=  final;
            *(Option->CallVegaError[ix]+it) +=  final*final;
            *(Option->CallRho[ix]+it) += rho;
            *(Option->CallRhoError[ix]+it) +=  rho*rho;

         } else {
 	   *(Option->PutValue[ix]+it) += (-del) ;
	   *(Option->PutError[ix]+it) += del*del ;
           delf = -del*fdelta1;
           *(Option->PutDelta[ix]+it) +=  delf;
           *(Option->PutDeltaError[ix]+it) +=  delf*delf;
           *(Option->PutVega[ix]+it) +=  -final;
           *(Option->PutVegaError[ix]+it) +=  final*final;
           *(Option->PutRho[ix]+it) -= rho;
           *(Option->PutRhoError[ix]+it) +=  rho*rho;


         }
      }


      for(it=1; it<=nt; it++){

        logarithmPrice = -CurrentPath[it] + *(Option->LogPrice[ir] + it);

        PriceWithDrift = exp( logarithmPrice );
 
        del = (PriceWithDrift- StrikePrice[ix]);
 
        final = -CurrentPath[it] - CurrentPath[0] - Var[ir]*Var[ir]*it;
        final *= PriceWithDrift/Var[ir];
        rho = 0.01*StrikePrice[ix]*(REAL)it;

         if( del >= 0.0){
            *(Option->CallValue[ix]+it) +=  del;
            *(Option->CallError[ix]+it) +=  del*del;
            delf = del*fdelta2;
            *(Option->CallDelta[ix]+it) +=  delf;
            *(Option->CallDeltaError[ix]+it) +=  delf*delf;
            *(Option->CallVega[ix]+it) +=  final;
            *(Option->CallVegaError[ix]+it) +=  final*final;
            *(Option->CallRho[ix]+it) += rho;
            *(Option->CallRhoError[ix]+it) +=  rho*rho;

         } else {
            *(Option->PutValue[ix]+it) += (-del) ;
            *(Option->PutError[ix]+it) += del*del ;
            delf = -del*fdelta2;
            *(Option->PutDelta[ix]+it) +=  delf;
            *(Option->PutDeltaError[ix]+it) +=  delf*delf;
            *(Option->PutVega[ix]+it) +=  -final;
            *(Option->PutVegaError[ix]+it) +=  final*final;
            *(Option->PutRho[ix]+it) -= rho;
            *(Option->PutRhoError[ix]+it) +=  rho*rho;


         }

      }

  } /* end of loop over strikes */

  V_Free((void *) Var);

}



void PrintOptionValues
                     (
                     EU_OPTION *Option
                     )
{
  int ir, it;
  int i;
  int nov = Option->DimOfOption;
  int NumTimeSteps = Option->TimeGridSize;
  REAL *StrikePrice = Option->Strike;

#ifdef MPI
  extern int NumberofProcessors, MyProcessorNumber;
#endif
#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif

  printf("\n\nEuropean call option values:\n");
#ifdef MPI
  printf("Reduced results\n");
#endif
  for(ir=1; ir<=nov; ir++){
    printf("Strike price =  %f\n", StrikePrice[ir]);
    for(it=0; it<=NumTimeSteps;it++)
      printf("Time slice #%d: %g %g %g %g %g %g %g %g\n", it, (*(Option->CallValue[ir]+it)), 
                                                  (*(Option->CallError[ir]+it)),
                                                  (*(Option->CallDelta[ir]+it)),
                                                  (*(Option->CallDeltaError[ir]+it)),  
                                                  (*(Option->CallVega[ir]+it)),
                                                  (*(Option->CallVegaError[ir]+it)),
                                                  (*(Option->CallRho[ir]+it)),
                                                  (*(Option->CallRhoError[ir]+it))  
                                                  );
  }

  printf("\n\nEuropean put option values:\n");
#ifdef MPI
  printf("Reduced results\n");
#endif
   for(ir=1; ir<=nov; ir++){
    printf("Strike price =  %f\n", StrikePrice[ir]);
    for(it=0; it<=NumTimeSteps;it++)
      printf("Time slice #%d: %g %g %g %g %g %g %g %g\n", it, (*(Option->PutValue[ir]+it)), 
                                                  (*(Option->PutError[ir]+it)),
                                                  (*(Option->PutDelta[ir]+it)),
                                                  (*(Option->PutDeltaError[ir]+it)),  
                                                  (*(Option->PutVega[ir]+it)),
                                                  (*(Option->PutVegaError[ir]+it)), 
                                                  (*(Option->PutRho[ir]+it)),
                                                  (*(Option->PutRhoError[ir]+it)) 
                                                  );
  }

#ifdef MPI
 }
#endif
}



#ifdef MPI

void GatherResults
                 (
                 EU_OPTION *Option
                 )
{
  REAL **CallPayoffBuffer = NULL;
  REAL **PutPayoffBuffer = NULL;
  REAL **CallErrorBuffer = NULL;
  REAL **PutErrorBuffer = NULL;

  REAL **CallDeltaBuffer = NULL;
  REAL **PutDeltaBuffer = NULL;
  REAL **CallDeltaErrorBuffer = NULL;
  REAL **PutDeltaErrorBuffer = NULL;

  REAL **CallVegaBuffer = NULL;
  REAL **PutVegaBuffer = NULL;
  REAL **CallVegaErrorBuffer = NULL;
  REAL **PutVegaErrorBuffer = NULL;

  REAL **CallRhoBuffer = NULL;
  REAL **PutRhoBuffer = NULL;
  REAL **CallRhoErrorBuffer = NULL;
  REAL **PutRhoErrorBuffer = NULL;
  int i; 
  int n; 
  int ir; 
  int it; 
  int size;
  int nov = Option->DimOfOption;
  int nt = Option->NumOptionTimeValues;


    size = (nov+1)*sizeof(REAL *);
    
    CallPayoffBuffer = (REAL **) mymalloc(size, "GatherResults(): CallPayoffBuffer\n");
    PutPayoffBuffer = (REAL **) mymalloc(size,"GatherResults(): PutPayoffBuffer\n");
    CallErrorBuffer = (REAL **) mymalloc(size, "GatherResults(): CallErrorBuffer\n");
    PutErrorBuffer = (REAL **) mymalloc(size,"GatherResults(): PutErrorBuffer\n");

    CallDeltaBuffer = (REAL **) mymalloc(size, "GatherResults(): CallDeltaBuffer\n");
    PutDeltaBuffer = (REAL **) mymalloc(size,"GatherResults(): PutDEltaBuffer\n");
    CallDeltaErrorBuffer = (REAL **) mymalloc(size, "GatherResults(): CallDeltaErrorBuffer\n");
    PutDeltaErrorBuffer = (REAL **) mymalloc(size,"GatherResults(): PutDeltaErrorBuffer\n");

    CallVegaBuffer = (REAL **) mymalloc(size, "GatherResults(): CallVegaBuffer\n");
    PutVegaBuffer = (REAL **) mymalloc(size,"GatherResults(): PutVegaBuffer\n");
    CallVegaErrorBuffer = (REAL **) mymalloc(size, "GatherResults(): CallVegaErrorBuffer\n");
    PutVegaErrorBuffer = (REAL **) mymalloc(size,"GatherResults(): PutVegaErrorBuffer\n");
    
    CallRhoBuffer = (REAL **) mymalloc(size, "GatherResults(): CallRhoBuffer\n");
    PutRhoBuffer = (REAL **) mymalloc(size,"GatherResults(): PutRhoBuffer\n");
    CallRhoErrorBuffer = (REAL **) mymalloc(size, "GatherResults(): CallRhoErrorBuffer\n");
    PutRhoErrorBuffer = (REAL **) mymalloc(size,"GatherResults(): PutRhoErrorBuffer\n");
    
    CallPayoffBuffer[0]=(REAL *) mymalloc(sizeof(REAL),"GatherResults():CallPayoffBuffer[0]\n");
    PutPayoffBuffer[0]=(REAL *) mymalloc(sizeof(REAL),"GatherResults():PutPayoffBuffer[0]\n");
    CallErrorBuffer[0]=(REAL *) mymalloc(sizeof(REAL),"GatherResults():CallErrorBuffer[0]\n");
    PutErrorBuffer[0]=(REAL *) mymalloc(sizeof(REAL),"GatherResults():PutErrorBuffer[0]\n");

    CallDeltaBuffer[0] = (REAL *) mymalloc(sizeof(REAL), "GatherResults(): CallDeltaBuffer[0]\n");
    PutDeltaBuffer[0] = (REAL *) mymalloc(sizeof(REAL),"GatherResults(): PutDEltaBuffer[0]\n");
    CallDeltaErrorBuffer[0] = (REAL *) mymalloc(sizeof(REAL), "GatherResults(): CallDeltaErrorBuffer[0]\n");
    PutDeltaErrorBuffer[0] = (REAL *) mymalloc(sizeof(REAL),"GatherResults(): PutDeltaErrorBuffer[0]\n");

    CallVegaBuffer[0] = (REAL *) mymalloc(sizeof(REAL), "GatherResults(): CallVegaBuffer[0]\n");
    PutVegaBuffer[0] = (REAL *) mymalloc(sizeof(REAL),"GatherResults(): PutVegaBuffer[0]\n");
    CallVegaErrorBuffer[0] = (REAL *) mymalloc(sizeof(REAL), "GatherResults(): CallVegaErrorBuffer[0]\n");
    PutVegaErrorBuffer[0] = (REAL *) mymalloc(sizeof(REAL),"GatherResults(): PutVegaErrorBuffer[0]\n");

    CallRhoBuffer[0] = (REAL *) mymalloc(sizeof(REAL), "GatherResults(): CallRhoBuffer[0]\n");
    PutRhoBuffer[0] = (REAL *) mymalloc(sizeof(REAL),"GatherResults(): PutRhoBuffer[0]\n");
    CallRhoErrorBuffer[0] = (REAL *) mymalloc(sizeof(REAL), "GatherResults(): CallRhoErrorBuffer[0]\n");
    PutRhoErrorBuffer[0] = (REAL *) mymalloc(sizeof(REAL),"GatherResults(): PutRhoErrorBuffer[0]\n");


    size = (nt+1)*sizeof(REAL);
    for(ir=1;ir<=nov;ir++){
        CallPayoffBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): CallPayoffBuffer\n");
        PutPayoffBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): PutPayoffBuffer\n");
        CallErrorBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): CallErrorBuffer\n");
        PutErrorBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): PutErrorBuffer\n");

        CallDeltaBuffer[ir] = (REAL *) mymalloc(size, "GatherResults(): CallDeltaBuffer\n");
        PutDeltaBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): PutDEltaBuffer\n");
        CallDeltaErrorBuffer[ir] = (REAL *) mymalloc(size, "GatherResults(): CallDeltaErrorBuffer\n");
        PutDeltaErrorBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): PutDeltaErrorBuffer\n");

        CallVegaBuffer[ir] = (REAL *) mymalloc(size, "GatherResults(): CallVegaBuffer\n");
        PutVegaBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): PutDEltaBuffer\n");
        CallVegaErrorBuffer[ir] = (REAL *) mymalloc(size, "GatherResults(): CallVegaErrorBuffer\n");
        PutVegaErrorBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): PutVegaErrorBuffer\n");

        CallRhoBuffer[ir] = (REAL *) mymalloc(size, "GatherResults(): CallRhoBuffer\n");
        PutRhoBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): PutDEltaBuffer\n");
        CallRhoErrorBuffer[ir] = (REAL *) mymalloc(size, "GatherResults(): CallRhoErrorBuffer\n");
        PutRhoErrorBuffer[ir] = (REAL *) mymalloc(size,"GatherResults(): PutRhoErrorBuffer\n");
    }


  for(ir=1;ir<=nov;ir++){
    MPI_Reduce(Option->CallValue[ir], CallPayoffBuffer[ir] , nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->PutValue[ir], PutPayoffBuffer[ir], nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->CallError[ir], CallErrorBuffer[ir] , nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->PutError[ir], PutErrorBuffer[ir], nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);

    MPI_Reduce(Option->CallDelta[ir], CallDeltaBuffer[ir] , nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->PutDelta[ir], PutDeltaBuffer[ir], nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->CallDeltaError[ir], CallDeltaErrorBuffer[ir] , nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->PutDeltaError[ir], PutDeltaErrorBuffer[ir], nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);

    MPI_Reduce(Option->CallVega[ir], CallVegaBuffer[ir] , nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->PutVega[ir], PutVegaBuffer[ir], nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->CallVegaError[ir], CallVegaErrorBuffer[ir] , nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->PutVegaError[ir], PutVegaErrorBuffer[ir], nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);

    MPI_Reduce(Option->CallRho[ir], CallRhoBuffer[ir] , nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->PutRho[ir], PutRhoBuffer[ir], nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->CallRhoError[ir], CallRhoErrorBuffer[ir] , nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
    MPI_Reduce(Option->PutRhoError[ir], PutRhoErrorBuffer[ir], nt+1 , MPI_REAL, MPI_SUM, 0, MPI_COMM_WORLD);
  }

  if(MyProcessorNumber == MASTER) {
    for(ir=1;ir<=nov;ir++){
      for(it=0; it<=nt; it++) {
        *(Option->CallValue[ir] +it) =  (*(CallPayoffBuffer[ir]+it))/NumberofProcessors;
        *(Option->PutValue[ir] +it) =  (*(PutPayoffBuffer[ir]+it))/NumberofProcessors;
        *(Option->CallError[ir] +it) = ( *(CallErrorBuffer[ir]+it))/NumberofProcessors;
        *(Option->PutError[ir] +it) =  (*(PutErrorBuffer[ir]+it))/NumberofProcessors;

        *(Option->CallDelta[ir] +it) =  (*(CallDeltaBuffer[ir]+it))/NumberofProcessors;
        *(Option->PutDelta[ir] +it) =  (*(PutDeltaBuffer[ir]+it))/NumberofProcessors;
        *(Option->CallDeltaError[ir] +it) = ( *(CallDeltaErrorBuffer[ir]+it))/NumberofProcessors;
        *(Option->PutDeltaError[ir] +it) =  (*(PutDeltaErrorBuffer[ir]+it))/NumberofProcessors;

        *(Option->CallVega[ir] +it) =  (*(CallVegaBuffer[ir]+it))/NumberofProcessors;
        *(Option->PutVega[ir] +it) =  (*(PutVegaBuffer[ir]+it))/NumberofProcessors;
        *(Option->CallVegaError[ir] +it) = ( *(CallVegaErrorBuffer[ir]+it))/NumberofProcessors;
        *(Option->PutVegaError[ir] +it) =  (*(PutVegaErrorBuffer[ir]+it))/NumberofProcessors;

        *(Option->CallRho[ir] +it) =  (*(CallRhoBuffer[ir]+it))/NumberofProcessors;
        *(Option->PutRho[ir] +it) =  (*(PutRhoBuffer[ir]+it))/NumberofProcessors;
        *(Option->CallRhoError[ir] +it) = ( *(CallRhoErrorBuffer[ir]+it))/NumberofProcessors;
        *(Option->PutRhoError[ir] +it) =  (*(PutRhoErrorBuffer[ir]+it))/NumberofProcessors;
      }
    }
  }

  M_Free((void **) CallPayoffBuffer, nov);
  M_Free((void **) PutPayoffBuffer, nov);
  M_Free((void **) CallErrorBuffer, nov);
  M_Free((void **) PutErrorBuffer, nov);

  M_Free((void **) CallDeltaBuffer, nov);
  M_Free((void **) PutDeltaBuffer, nov);
  M_Free((void **) CallDeltaErrorBuffer, nov);
  M_Free((void **) PutDeltaErrorBuffer, nov);

  M_Free((void **) CallVegaBuffer, nov);
  M_Free((void **) PutVegaBuffer, nov);
  M_Free((void **) CallVegaErrorBuffer, nov);
  M_Free((void **) PutVegaErrorBuffer, nov);

  M_Free((void **) CallRhoBuffer, nov);
  M_Free((void **) PutRhoBuffer, nov);
  M_Free((void **) CallRhoErrorBuffer, nov);
  M_Free((void **) PutRhoErrorBuffer, nov);

}

#endif


void EU_OPTION_Free(EU_OPTION *Option)
{
  int NumTimeSteps = Option->TimeGridSize;
  int nt = Option->NumOptionTimeValues;
  int dof = Option->DimOfUnderlying;
  int nov = Option->DimOfOption;

  if(Option == NULL) {
    fprintf(stderr,"EU_OPTION_Free(): cannot free a null option\n");
    exit(1);
  }
  Option->Underlying = NULL;
  Option->PayoffFunction = NULL;
  Option->Strike = NULL;

  V_Free((void *) Option->InitPrice);
  if(Option->ValuationTimeIndex != NULL) V_Free((void *) Option->ValuationTimeIndex);
  V_Free((void *) Option->InterestRate);
  V_Free((void *) Option->Drift);
  V_Free((void *) Option->DiscountBeg);
  V_Free((void *) Option->DiscountEnd);
  if(Option-> IntPayoffParams != NULL) V_Free((void *) Option->IntPayoffParams);
  V_Free((void *) Option->FloatPayoffParams);

  M_Free((void **) Option->LogPrice, dof);

  M_Free((void **) Option->CallValue, nov);
  M_Free((void **) Option->PutValue, nov);
  M_Free((void **) Option->CallError, nov);
  M_Free((void **) Option->PutError, nov);

  M_Free((void **) Option->CallDelta, nov);
  M_Free((void **) Option->PutDelta, nov);
  M_Free((void **) Option->CallDeltaError, nov);
  M_Free((void **) Option->PutDeltaError, nov);

  M_Free((void **) Option->CallRho, nov);

  M_Free((void **) Option->CallRhoError, nov);
  M_Free((void **) Option->CallVega, nov);
  M_Free((void **) Option->CallVegaError, nov);
  M_Free((void **) Option->PutRho, nov);
  M_Free((void **) Option->PutRhoError, nov);


  free((void *) Option);

}


int Externalize
              (
              int NumOptions,
              EU_OPTION **Options,
              char *OptionContainer
              )
{
  int ix, it;
  int nov;
  int NumTimeSteps;
  REAL iPrice;
  EU_OPTION *Option=NULL;
  REAL *StrikePrice = NULL;
  FILE *optfp=NULL;
  int iopt;


#ifdef MPI
  extern int NumberofProcessors, MyProcessorNumber;
#endif
#ifdef MPI
  if(MyProcessorNumber == MASTER) {
#endif
  if((optfp = fopen(OptionContainer,"w")) == NULL) {
        fprintf(stderr, "Externalize(): can't open option container\n");
        return -1;
  }


  for(iopt=1; iopt<=NumOptions; iopt++) {

      Option = Options[iopt];
      iPrice = *(Option->InitPrice + iopt);
      StrikePrice = Option->Strike;
      nov = Option->DimOfOption;
      NumTimeSteps = Option->NumOptionTimeValues;

      for(ix=1; ix<=nov; ix++){
         printf("Option %d %f %f\n", NumTimeSteps,iPrice,StrikePrice[ix]); 
        fprintf(optfp,"Option %d %f %f\n", NumTimeSteps,iPrice,StrikePrice[ix]); 
        for(it=0; it<=NumTimeSteps;it++) {
          fprintf(optfp,"%d %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f\n", 
                                                  it, 
                                                  (*(Option->CallValue[ix]+it)),
                                                  (*(Option->CallError[ix]+it)),
                                                  (*(Option->CallDelta[ix]+it)),
                                                  (*(Option->CallDeltaError[ix]+it)),
                                                  (*(Option->CallVega[ix]+it)),
                                                  (*(Option->CallVegaError[ix]+it)),
                                                  (*(Option->CallRho[ix]+it)),
                                                  (*(Option->CallRhoError[ix]+it)),
                                                  (*(Option->PutValue[ix]+it)),
                                                  (*(Option->PutError[ix]+it)),
                                                  (*(Option->PutDelta[ix]+it)),
                                                  (*(Option->PutDeltaError[ix]+it)),
                                                  (*(Option->PutVega[ix]+it)),
                                                  (*(Option->PutVegaError[ix]+it)),
                                                  (*(Option->PutRho[ix]+it)),
                                                  (*(Option->PutRhoError[ix]+it))
                                                  );
          printf("%d %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f\n", 
                                                  it, 
                                                  (*(Option->CallValue[ix]+it)),
                                                  (*(Option->CallError[ix]+it)),
                                                  (*(Option->CallDelta[ix]+it)),
                                                  (*(Option->CallDeltaError[ix]+it)),
                                                  (*(Option->CallVega[ix]+it)),
                                                  (*(Option->CallVegaError[ix]+it)),
                                                  (*(Option->CallRho[ix]+it)),
                                                  (*(Option->CallRhoError[ix]+it)),
                                                  (*(Option->PutValue[ix]+it)),
                                                  (*(Option->PutError[ix]+it)),
                                                  (*(Option->PutDelta[ix]+it)),
                                                  (*(Option->PutDeltaError[ix]+it)),
                                                  (*(Option->PutVega[ix]+it)),
                                                  (*(Option->PutVegaError[ix]+it)),
                                                  (*(Option->PutRho[ix]+it)),
                                                  (*(Option->PutRhoError[ix]+it))
                                                  );
		} /* end of the print for loop */
     } /* loop over strikes */

  } /* loop over init stock values */

  close(optfp);
  printf ("Done\n");
  return (0); 
#ifdef MPI
  }
#endif
  return (0); 
}


